%%
% This demo code accompanies the paper:
%
% "Dense non-rigid shape correspondence using random forests"
% E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
% Proc. CVPR 2014
%
% Please cite the paper above if you use the code.
%
% Emanuele Rodola (c) 2015
%
clc
close all
clear all

addpath('./tools/');
addpath(genpath('./manopt/manopt/'))

% NOTES:
% * This code will save temporary files in the shapes folder. Make sure you
%   have enough free space (tens of MB, depending on shape size).
% * load_training_shape() needs ground-truth labels. Look into the code if
%   you are not using TOSCA, KIDS, or FAUST datasets.

params = struct;
params.n_eigen = 100;         % the larger the better, but takes up RAM.
                              % recommended = 300 or more
params.n_energy_levels = 100; % WKS size
                              % recommended = 300
params.n_trees = 2;           % recommended = 15 or more
params.n_labels = 27894;      % assuming the training shapes have the same
                              % number of points n, this is equal to n

% If you want to recompile the mex code (uses BOOST):
% cd src;
% mex -v train_tree.cpp ShapeRFClassifier.cpp -I../../../boost_1_54_0
% mex -v predict.cpp ShapeRFClassifier.cpp -I../../../boost_1_54_0
% cd ..

% Only two training shapes are used here for demonstration purposes.
% Of course we do not expect any good results with such a training set.

train_shapes = {'./shapes/cat0', './shapes/cat1'};

trees = cell(1,params.n_trees);
for i=1:params.n_trees
    trees{i} = strcat('./tree', num2str(i), '.dat');
end

% -------------------------------------------------------------------------
% TRAIN
% -------------------------------------------------------------------------

for i=1:length(train_shapes)
    S = load_training_shape(train_shapes{i}, params);
    assert(S.n==params.n_labels)
end

for i=1:params.n_trees
    train_tree(train_shapes, params.n_eigen, params.n_energy_levels, trees{i});
end

% -------------------------------------------------------------------------
% TEST
% -------------------------------------------------------------------------

% Sanity check: all training shapes must get the ground-truth as a
% prediction.

for i=1:length(train_shapes)
    [r,c,v] = predict(train_shapes{i}, trees, params.n_eigen, params.n_energy_levels, 0, params.n_labels);
    T = sparse(r, c, v, params.n_labels, params.n_labels);
    assert(~any(diag(T-speye(params.n_labels))))
end

% Match two new shapes. Note that these must not necessarily have the same 
% number of vertices as the training shapes.

M = load_shape('./shapes/cat4', params);
tic;
[r,c,v] = predict('./shapes/cat4', trees, params.n_eigen, params.n_energy_levels, 0, params.n_labels);
t = toc;
fprintf('Random forest prediction took %.2f sec.\n', t);
T1 = sparse(r, c, v, M.n, params.n_labels);

N = load_shape('./shapes/cat5', params);
tic;
[r,c,v] = predict('./shapes/cat5', trees, params.n_eigen, params.n_energy_levels, 0, params.n_labels);
t = toc;
fprintf('Random forest prediction took %.2f sec.\n', t);
T2 = sparse(r, c, v, N.n, params.n_labels);

% T1 and T2 are stochastic matrices (each row sums up to 1).

% We now pass to the functional map representation and show the predicted
% matches.

C_F = (M.evecs' * M.S * T1) * (T2' * N.evecs);
[C_F, matches] = run_icp(M, N, params.n_eigen, C_F, 20);

figure
colors = create_colormap(M,M);
colormap(colors), plot_scalar_map(M, 1:M.n), axis off

figure
N2 = M; N2.VERT = M.VERT(matches,:);
colors = create_colormap(N2,M);
colormap(colors), plot_scalar_map(N, 1:N.n), axis off
title('Before regularization')

% We finally regularize the prediction.

[C, matches] = regularize(N, M, C_F, matches, 1e3, params);

figure
N2 = M; N2.VERT = M.VERT(matches,:);
colors = create_colormap(N2,M);
colormap(colors), plot_scalar_map(N, 1:N.n), axis off
title('After regularization')
